/*
 * $Id$
 *
    Copyright (c) 2006-2014 Chung, Hyung-Hwan. All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    1. Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _QSE_SI_SOCKET_HPP_
#define _QSE_SI_SOCKET_HPP_

#include <qse/Types.hpp>
#include <qse/Uncopyable.hpp>
#include <qse/si/SocketAddress.hpp>
#include <qse/si/sck.h>


/////////////////////////////////
QSE_BEGIN_NAMESPACE(QSE)
/////////////////////////////////

class Socket: public Uncopyable, public Types
{
public:
	enum ErrorCode
	{
		E_ENOERR,  /**< no error */
		E_EOTHER,  /**< other error */
		E_ENOIMPL, /**< not implemented */
		E_ESYSERR, /**< subsystem error */
		E_EINTERN, /**< internal error */

		E_ENOMEM,
		E_EINVAL,
		E_EACCES,
		E_EPERM,
		E_ENOENT,
		E_EEXIST,
		E_ENOTDIR,
		E_EINTR,
		E_EPIPE,
		E_EINPROG, /* in progress */
		E_EAGAIN  /* resource unavailable unavailable */
	};

	enum Trait
	{
		T_NONBLOCK = (1 << 0),
		T_CLOEXEC  = (1 << 1)
	};

	Socket () QSE_CPP_NOEXCEPT;
	~Socket () QSE_CPP_NOEXCEPT;

	ErrorCode getErrorCode () const QSE_CPP_NOEXCEPT { return this->errcode; }
	void setErrorCode (ErrorCode errcode) QSE_CPP_NOEXCEPT { this->errcode = errcode; }

	int open (int domain, int type, int protocol, int traits = 0) QSE_CPP_NOEXCEPT;
	void close () QSE_CPP_NOEXCEPT;

	int connect (const SocketAddress& target) QSE_CPP_NOEXCEPT;
	int bind (const SocketAddress& target) QSE_CPP_NOEXCEPT;
	int listen (int backlog) QSE_CPP_NOEXCEPT;
	int accept (Socket* newsck, SocketAddress* newaddr, int traits = 0) QSE_CPP_NOEXCEPT;

	// The send() functions sends data by attemping a single call to the 
	// underlying system calls
	qse_ssize_t send (const void* buf, qse_size_t len) QSE_CPP_NOEXCEPT;
	qse_ssize_t send (const void* buf, qse_size_t len, const SocketAddress& dstaddr) QSE_CPP_NOEXCEPT;

	// The sendx() functions sends data as much as it can, possibly with multiple
	// underlying system calls. 
	int sendx (const void* buf, qse_size_t len, qse_size_t* total_sent = QSE_NULL) QSE_CPP_NOEXCEPT;
	int sendx (const void* buf, qse_size_t len, const SocketAddress& dstaddr, qse_size_t* total_sent = QSE_NULL) QSE_CPP_NOEXCEPT;
	int sendx (qse_ioptl_t* vec, int count, qse_size_t* total_sent = QSE_NULL) QSE_CPP_NOEXCEPT;

	qse_ssize_t receive (void* buf, qse_size_t len) QSE_CPP_NOEXCEPT;
	qse_ssize_t receive (void* buf, qse_size_t len, SocketAddress& srcaddr) QSE_CPP_NOEXCEPT;

/* TODO: sendmsg, recvmsg */
protected:
	qse_sck_hnd_t handle;
	ErrorCode errcode;

	void set_errcode_with_syserr (int syserr);
};


/////////////////////////////////
QSE_END_NAMESPACE(QSE)
/////////////////////////////////


#endif
